#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <unistd.h>
#include <time.h>
#include <assert.h>

#include <jack/jack.h>

#include "fourier.h"
#include "plot.h"
#include "types.h"
#include "sound.h"
#include "complex.h"
#include "window.h"

static real cache_in[N/2];
static real cache_out[N/2];

static jack_client_t *client;

static jack_port_t *input_port;

static jack_port_t *output_portL;
static jack_port_t *output_portR;

static int process (jack_nframes_t nframes, void *arg);

static void
sound_jack_open (void)
{
  bzero (cache_in, N/2*sizeof(real));
  bzero (cache_out, N/2*sizeof(real));

  const char **ports;
  const char *physical_in_port, *physical_out_portL, *physical_out_portR;

  /* try to become a client of the JACK server */

  if ((client = jack_client_new ("robot")) == 0)
    {
      fprintf (stderr, "jack server not running?\n");
      exit (EXIT_FAILURE);
  }

  if (jack_get_sample_rate (client) != SAMPLING_RATE)
    {
      fprintf (stderr, "sampling rate error!\n");
      exit (EXIT_FAILURE);
    }

  /* set up the callback function */

  jack_set_process_callback (client, process, 0);

  /* create ports */

  input_port = jack_port_register (client, "input", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);

  output_portL = jack_port_register (client, "output-left", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
  output_portR = jack_port_register (client, "output-right", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);

  /* lauch the client */

  if (jack_activate (client))
    {
      fprintf (stderr, "cannot activate client.\n");
      exit (EXIT_FAILURE);
    }

  /* get the physical ports */

  if ((ports = jack_get_ports (client, NULL, NULL, JackPortIsPhysical|JackPortIsOutput)) == NULL)
    {
      fprintf (stderr, "cannot find any physical recording port.\n");
      exit (EXIT_FAILURE);
    }

  if (ports[0] != NULL)
    {
      physical_in_port = ports[0];
    }
  else
    {
      fprintf (stderr, "physical input port not found.\n");
      exit (EXIT_FAILURE);
    }

  free (ports);

  /* connect to the ports */

  if (jack_connect (client, physical_in_port, jack_port_name (input_port)))
    {
      fprintf (stderr, "cannot connect input ports...\n");
    }

  /* get the physical ports */

  if ((ports = jack_get_ports (client, NULL, NULL, JackPortIsPhysical|JackPortIsInput)) == NULL)
    {
      fprintf (stderr, "cannot find any physical playback port.\n");
      exit (EXIT_FAILURE);
    }

  if (ports[0] != NULL)
    {
      physical_out_portL = ports[0];
    }
  else
    {
      fprintf (stderr, "left physical output port not found.\n");
      exit (EXIT_FAILURE);
    }

  if (ports[1] != NULL)
    {
      physical_out_portR = ports[1];
    }
  else
    {
      fprintf (stderr, "right physical output port not found.\n");
      exit (EXIT_FAILURE);
    }

  free (ports);

  /* connect to the ports */

  if (jack_connect (client, jack_port_name (output_portL), physical_out_portL) ||
      jack_connect (client, jack_port_name (output_portR), physical_out_portR))
    {
      fprintf (stderr, "cannot connect output ports...\n");
    }
}

static void
sound_jack_close (void)
{
  jack_client_close (client);
}

static void
sound_jack_read (frame s, jack_default_audio_sample_t *data)
{
  int i;

  assert (s && data);

  for (i=0; i<(N/2); i++)
    {
      s[i] = cache_in[i];
      s[i+N/2] = cache_in[i] = *data++;
    }
}

static void
sound_jack_write (jack_default_audio_sample_t *data, frame s)
{
  int i;

  assert (data && s);

  for (i=0; i<(N/2); i++)
    {
      *data++ = s[i] + cache_out[i];
      cache_out[i] = s[N/2+i];
    }
}

static void sound_transform (frame s);

static int
process (jack_nframes_t nframes, void *arg)
{
  static frame s;

  if (nframes != (N/2)) return 1;

  {
    jack_default_audio_sample_t *in = (jack_default_audio_sample_t *) jack_port_get_buffer (input_port, nframes);

    jack_default_audio_sample_t *outL = (jack_default_audio_sample_t *) jack_port_get_buffer (output_portL, nframes);
    jack_default_audio_sample_t *outR = (jack_default_audio_sample_t *) jack_port_get_buffer (output_portR, nframes);

    sound_jack_read (s, in);

    sound_transform (s);

    sound_jack_write (outL, s);
    sound_jack_write (outR, s);
  }

  return 0;      
}

static void
sound_transform (frame s)
{
  static spectrum S;

  static real amp[N], phs[N];

  int i;

  real sum;

  /* Fourier analysis */

  sum = hann_window (s);
  zero_phase (s);

  fft (s, S);

  cartesian_to_polar (S, amp, phs);
  amp_normalize (amp, sum);
  plot_spectral (amp);

  /* transform: robotize! */

  for (i=0; i<N; i++)
    {
      phs[i] = 0;
    }

  /* Fourier synthesis */

  polar_to_cartesian (amp, phs, S);

  ifft (S, s);

  zero_phase (s);
}

int
main (int argc, char *argv[])
{
  assert (sizeof(short) == 2);

  /* init */

  plot_init ();
  fft_init ();

  sound_jack_open ();

  /* process */

  while (1)
    {
      sleep (60);
    }

  /* exit */

  sound_jack_close ();

  fft_exit ();
  plot_exit ();

  exit (EXIT_SUCCESS);

  return 0;
}

/* End Of File */
